{ This file was automatically created by Lazarus. do not edit!
  This source is only used to compile and install the package.
 }

unit dcp; 

interface

uses
  DCPbase64, DCPblockciphers, DCPconst, DCPcrypt2, DCPblowfish, DCPcast128, 
  DCPcast256, DCPdes, DCPgost, DCPice, DCPidea, DCPmars, DCPmisty1, DCPrc2, 
  DCPrc4, DCPrc5, DCPrc6, DCPrijndael, DCPserpent, DCPtea, DCPtwofish, 
  DCPhaval, DCPmd4, DCPmd5, DCPripemd128, DCPripemd160, DCPsha1, DCPsha256, 
  DCPsha512, DCPtiger;

implementation

end.
